#if ARDUINO >= 100
#include "Arduino.h"
#else
#include "WProgram.h"
#endif

#include "MemoryLCD.h"
#include <SPI.h>
#include "resources.h"

MemoryLCD::MemoryLCD(const uint8_t SCSpin, const uint8_t SIpin, const uint8_t SCLKpin, const uint8_t DISPpin, const uint8_t EXTCOMINpin, const uint8_t RAMCSpin, const boolean useEXTCOMIN, displayType d)
    : SCS(SCSpin),
      SI(SIpin),
      SCLK(SCLKpin),
      DISP(DISPpin),
      EXTCOMIN(EXTCOMINpin),
      RAMCS(RAMCSpin),
      enablePWM(useEXTCOMIN),
      colourDepth(d),
      EXTCOMIN_PIN_STATE(false),
      pwm_interrupt_counter(0),
      commandByte(0b10000000),
      vcomByte(0b01000000),
      savedSpiControlRegister(SPCR),
      sRam(RAMCS)
{
  // Delay for MemoryLCD to reach 5V
  delayMicroseconds(800); // minimum 750us
}

void MemoryLCD::begin(void)
{
  sRam.writeByte(0, 0xa5);
  uint8_t dat = sRam.readByte(0);
  Serial.println(dat, HEX);
  sRam.writeByte(0, 0xc8);
  dat = sRam.readByte(0);
  Serial.println(dat, HEX);

  // setup timer for EXTCOMIN signal
  if (enablePWM)
  {
    pinMode(EXTCOMIN, OUTPUT);
    digitalWrite(EXTCOMIN, LOW);
    // New interrupt based EXCOMIN toggle, ARDUINO SPECIFIC
    TIMSK0 |= (1 << OCIE0A);
  }
  else
  {
    // TODO: setup timer driven interrupt instead?
  }

  // setup SPI
  SPI.begin();
  // Datasheet says SPI clock must have <1MHz frequency (SPI_CLOCK_DIV16)
  // but it may work at 2MHz (SPI_CLOCK_DIV8) or even 4MHz (SPI_CLOCK_DIV4)
  lcdSpiSettings = SPISettings(SPI_CLOCK_DIV4, MSBFIRST, SPI_MODE0);
  // Set pin modes
  pinMode(SCS, OUTPUT);
  pinMode(SI, OUTPUT);
  pinMode(SCLK, OUTPUT);
  digitalWrite(SCS, LOW);
  digitalWrite(SI, LOW);
  digitalWrite(SCLK, LOW);
  if (DISP != 255)
  {
    pinMode(DISP, OUTPUT);
    digitalWrite(DISP, LOW);
  }
  if (enablePWM)
  {
    pinMode(EXTCOMIN, OUTPUT);
  }

  // Memory LCD startup sequence
  digitalWrite(DISP, HIGH);
  delayMicroseconds(PWRUP_DISP_DELAY);

  pinMode(EXTCOMIN, OUTPUT);
  delayMicroseconds(PWRUP_EXTCOMIN_DELAY);

  clearRam();
}

void MemoryLCD::end()
{
  if (enablePWM)
  {
    TIMSK0 |= (1 << OCIE0A); // TODO disable this bit! (this line just copied from begin())
  }
  SPI.end();
}

void MemoryLCD::clearRam(void)
{
  // initial sram
  for (uint8_t i = 0; i < LCDWIDTH / 8; i++)
  {
    lineBuffer[i] = 0xFF;
  }
  for (byte y = 0; y < LCDHEIGHT; y++)
  {
    sRam.writeBuffer(y * LCDWIDTH / 8, lineBuffer, LCDWIDTH / 8);
  }
}

void MemoryLCD::dispMultiLines(const uint8_t *const data, const uint8_t lineNumber, const uint8_t numLines)
{
  SPI.beginTransaction(lcdSpiSettings);
  // this implementation writes multiple lines that are CONSECUTIVE (although they don't have to be (if this
  // funcion was appropriately modified), as an address is given for every line, not just the first in the sequence)
  // data for all lines should be stored in a single array
  const uint8_t *linePtr = data;
  digitalWrite(SCS, HIGH);
  delayMicroseconds(SCS_HIGH_DELAY);
  SPI.transfer(0b10000000);
  for (byte y = lineNumber + 1; y < lineNumber + numLines + 1; y++)
  {
    SPI.transfer(reverseBitOrder(y)); // only the line number goes LSBFIRST for some reason
    for (byte x = 0; x < (LCDWIDTH * colourDepth) / 8; x++)
    {
      SPI.transfer(*(linePtr++));
    }
    SPI.transfer(0x00);
    linePtr = data;
  }
  SPI.transfer(0x00); // trailing padding chars
  delayMicroseconds(SCS_LOW_DELAY);
  digitalWrite(SCS, LOW);
  delayMicroseconds(INTERFRAME_DELAY);
  SPI.endTransaction();

  sRam.writeBuffer(LCDWIDTH / 8 * lineNumber, data, LCDWIDTH / 8 * numLines);
}

void MemoryLCD::drawPixel(uint16_t x, uint16_t y, uint32_t color)
{
  //sRam.readBuffer(LCDWIDTH / 8 * y, lineBuffer, LCDWIDTH / 8);
  modByte = sRam.readByte(LCDWIDTH / 8 * y + x/8);

  modBit = 7 - x % 8;

  if (color)
  {
    modByte &= ~(1 << modBit);
  }
  else
  {
    modByte |= (1 << modBit);
  }

  sRam.writeByte(LCDWIDTH / 8 * y + x/8, modByte);
}

void MemoryLCD::writeFrameBufferToDisplay()
{
  for (uint16_t y = 0; y < LCDHEIGHT; y++)
  {
    sRam.readBuffer(LCDWIDTH / 8 * y, lineBuffer, LCDWIDTH / 8);
    dispMultiLines(lineBuffer, y);
  }
}

void MemoryLCD::clearDisplay()
{
  SPI.beginTransaction(lcdSpiSettings);
  digitalWrite(SCS, HIGH);
  delayMicroseconds(SCS_HIGH_DELAY);
  SPI.transfer(0b00100000);
  SPI.transfer(0x00);
  delayMicroseconds(SCS_LOW_DELAY);
  digitalWrite(SCS, LOW);
  delayMicroseconds(INTERFRAME_DELAY);
  SPI.endTransaction();

  clearRam();
}

void MemoryLCD::drawChar1616(uint8_t x, uint8_t y, uint8_t chChar)
{
    uint8_t i, j;
    uint8_t chTemp = 0, y0 = y, chMode = 0;

    for (i = 0; i < 32; i ++) {
        chTemp = pgm_read_byte(&Font1612[chChar - 0x30][i]);
        for (j = 0; j < 8; j ++) {
            chMode = chTemp & 0x80? 1 : 0; 
            drawPixel(x, y, chMode);
            chTemp <<= 1;
            y ++;
            if ((y - y0) == 16) {
                y = y0;
                x ++;
                break;
            }
        }
    }
}

void MemoryLCD::drawChar(uint8_t x, uint8_t y, uint8_t acsii, uint8_t size, uint8_t mode)
{
    uint8_t i, j, y0=y;
    char temp;
    uint8_t ch = acsii - ' ';
    for(i = 0;i<size;i++) {
        if(size == 12)
        {
            if(mode)temp = pgm_read_byte(&Font1206[ch][i]);
            else temp = ~pgm_read_byte(&Font1206[ch][i]);
        }
        else 
        {            
            if(mode)temp = pgm_read_byte(&Font1608[ch][i]);
            else temp = ~pgm_read_byte(&Font1608[ch][i]);
        }
        for(j =0;j<8;j++)
        {
            if(temp & 0x80) drawPixel(x, y, 1);
            else drawPixel(x, y, 0);
            temp <<= 1;
            y++;
            if((y-y0) == size)
            {
                y = y0;
                x++;
                break;
            }
        }
    }
}

void MemoryLCD::drawString(uint8_t x, uint8_t y, const char *pString, uint8_t Size, uint8_t Mode)
{
    while (*pString != '\0') {       
        if (x > (LCDWIDTH - Size / 2)) {
            x = 0;
            y += Size;
            if (y > (LCDHEIGHT - Size)) {
                y = x = 0;
            }
        }
        
        char(x, y, *pString, Size, Mode);
        x += Size / 2;
        pString++;
    }
}

void MemoryLCD::drawChar3216(uint8_t x, uint8_t y, uint8_t chChar)
{
    uint8_t i, j;
    uint8_t chTemp = 0, y0 = y, chMode = 0;

    for (i = 0; i < 64; i++) {
        chTemp = pgm_read_byte(&Font3216[chChar - 0x30][i]);
        for (j = 0; j < 8; j++) {
            chMode = chTemp & 0x80? 1 : 0; 
            drawPixel(x, y, chMode);
            chTemp <<= 1;
            y++;
            if ((y - y0) == 32) {
                y = y0;
                x++;
                break;
            }
        }
    }
}

void MemoryLCD::drawBitmap(uint8_t x, uint8_t y, const uint8_t *pBmp, uint8_t chWidth, uint8_t chHeight)
{
    uint8_t i,j,byteWidth = (chWidth + 7)/8;
    for(j = 0;j < chHeight;j++){
        for(i = 0;i <chWidth;i ++){
            if(pgm_read_byte(pBmp +j*byteWidth+i/8) & (128 >> (i & 7))){
                drawPixel(x+i,y+j,1);
            }
        }
    }        
}

// won't work if DISP pin is not used
void MemoryLCD::turnOff()
{
  if (DISP != 255)
    digitalWrite(DISP, HIGH);
}

// won't work if DISP pin is not used
void MemoryLCD::turnOn()
{
  if (DISP != 255)
    digitalWrite(DISP, LOW);
}

int MemoryLCD::getLcdWidth() const
{
  return LCDWIDTH;
}

int MemoryLCD::getLcdHeight() const
{
  return LCDHEIGHT;
}

void MemoryLCD::softToggleVCOM()
{
  SPI.beginTransaction(lcdSpiSettings);
  vcomByte ^= (0b01000000);
  digitalWrite(SCS, HIGH);
  delayMicroseconds(SCS_HIGH_DELAY);
  SPI.transfer(vcomByte);
  SPI.transfer(0x00);
  delayMicroseconds(SCS_LOW_DELAY);
  digitalWrite(SCS, LOW);
  delayMicroseconds(10);
  SPI.endTransaction();
}

// Recommended to be used in conjunction with an interrupt service routine (see below hardToggleVCOM method for code)
void MemoryLCD::hardToggleVCOM()
{
  if (enablePWM)
  {
    if (++pwm_interrupt_counter > 243)
    { // 243 gives 2Hz EXTCOMIN rate
      EXTCOMIN_PIN_STATE = !EXTCOMIN_PIN_STATE;
      digitalWrite(EXTCOMIN, EXTCOMIN_PIN_STATE);
      pwm_interrupt_counter = 0;
    }
  }
}

const uint8_t MemoryLCD::reverseBitOrder(const uint8_t b)
{
  uint8_t temp = b;
  temp = (temp & 0xF0) >> 4 | (temp & 0x0F) << 4;
  temp = (temp & 0xCC) >> 2 | (temp & 0x33) << 2;
  temp = (temp & 0xAA) >> 1 | (temp & 0x55) << 1;
  return temp;
}
