#include "MemoryLCD.h"
#include <stdio.h>
#include <string.h>

extern TIM_HandleTypeDef htim6;
extern SPI_HandleTypeDef hspi1;

uint8_t commandByte = 0x80;
uint8_t vcomByte    = 0x40;
uint8_t clearByte   = 0x20;
uint8_t paddingByte = 0x00;
  
uint8_t lineBuffer[LCDWIDTH/8];
unsigned char frameBuffer[LCDWIDTH * LCDHEIGHT / 8];
uint8_t reverseByte(uint8_t b);

void delayMicroseconds(uint32_t us){
	us = us * 32;
	while(us--) {
	}
}

void spi_writenb(uint8_t *data, uint32_t length) {
	HAL_SPI_Transmit(&hspi1, data, length, 1000);
}

void initMemoryLCD(void) {
  delayMicroseconds(800);  // minimum 750us
  HAL_TIM_Base_Start_IT(&htim6);
  
  // Set pin modes
	CS(GPIO_LO);
	DISP(GPIO_LO);
	ECOM(GPIO_LO);
	
	// Memory LCD startup sequence with recommended timings
	DISP(GPIO_HI);
	delayMicroseconds(PWRUP_DISP_DELAY);
	
	ECOM(GPIO_LO);
	delayMicroseconds(PWRUP_EXTCOMIN_DELAY);
	
	clearLineBuffer();
}

void writeLineToDisplay(uint8_t lineNumber, uint8_t *line) {
	writeMultipleLinesToDisplay(lineNumber, 1, line);
}


void writeMultipleLinesToDisplay(uint8_t lineNumber, uint8_t numLines, uint8_t *lines) {
  // this implementation writes multiple lines that are CONSECUTIVE (although they don't 
  // have to be, as an address is given for every line, not just the first in the sequence)
  // data for all lines should be stored in a single array
  uint8_t * linePtr = lines;

	CS(GPIO_HI);
  delayMicroseconds(SCS_HIGH_DELAY);
  spi_writenb(&commandByte, 1);
  for(char x=0; x<numLines; x++) {
    uint8_t reversedLineNumber = reverseByte(lineNumber);
    spi_writenb(&reversedLineNumber, 1);
	// Transfers a whole line of data at once
	// pointer arithmetic assumes an array of lines - TODO: change this?
	spi_writenb(linePtr+(LCDWIDTH/8)*x, LCDWIDTH/8);
    spi_writenb(&paddingByte, 1);
    lineNumber++;
  }
  spi_writenb(&paddingByte, 1);  // trailing paddings
  delayMicroseconds(SCS_LOW_DELAY);

  CS(GPIO_LO);
  delayMicroseconds(INTERFRAME_DELAY);	// can I delete this delay?
}


void writePixelToLineBuffer(unsigned int pixel, uint8_t isWhite) {
// pixel location expected in the fn args follows the scheme defined in the datasheet.
// NB: the datasheet defines pixel addresses starting from 1, NOT 0
  if((pixel <= LCDWIDTH) && (pixel != 0)) {
    pixel = pixel - 1;
    if(isWhite)
      lineBuffer[pixel/8] |=  (1 << (7 - pixel%8));
    else
      lineBuffer[pixel/8] &= ~(1 << (7 - pixel%8));
  }  
}


void writeByteToLineBuffer(uint8_t byteNumber, uint8_t byteToWrite) {
// char location expected in the fn args has been extrapolated from the pixel location
// format (see above), so chars go from 1 to LCDWIDTH/8, not from 0
  if(byteNumber <= LCDWIDTH/8 && byteNumber != 0) {
    byteNumber -= 1;
    lineBuffer[byteNumber] = byteToWrite;
  } 
}


void copyByteWithinLineBuffer(uint8_t sourceByte, uint8_t destinationByte) {
  if(sourceByte <= LCDWIDTH/8 && destinationByte <= LCDWIDTH/8) {
    sourceByte -= 1;
    destinationByte -= 1;
    lineBuffer[destinationByte] = lineBuffer[sourceByte];
  } 
}


void setLineBufferBlack(void) {
  for(char i=0; i<LCDWIDTH/8; i++) {
    lineBuffer[i] = 0x00;
  }
}


void setLineBufferWhite(void) {
  for(char i=0; i<LCDWIDTH/8; i++) {
    lineBuffer[i] = 0xFF;
  }
}

void writeLineBufferToDisplay(uint8_t lineNumber) {
  writeMultipleLinesToDisplay(lineNumber, 1, lineBuffer);
}


void writeLineBufferToDisplayRepeatedly(uint8_t lineNumber, uint8_t numLines) {
  writeMultipleLinesToDisplay(lineNumber, numLines, lineBuffer);
}


void writePixelToFrameBuffer(uint16_t pixel, uint16_t lineNumber, uint8_t isWhite) {
// pixel location expected in the fn args follows the scheme defined in the datasheet.
// NB: the datasheet defines pixel addresses starting from 1, NOT 0
  if((pixel <= LCDWIDTH) && (pixel != 0) && (lineNumber <=LCDHEIGHT) & (lineNumber != 0)) {
    pixel -= 1;
    lineNumber -= 1;
    if(isWhite)
      frameBuffer[(lineNumber*LCDWIDTH/8)+(pixel/8)] |=  (1 << (7 - pixel%8));
    else
      frameBuffer[(lineNumber*LCDWIDTH/8)+(pixel/8)] &= ~(1 << (7 - pixel%8));
  }  
}


void writeByteToFrameBuffer(uint8_t byteNumber, uint8_t lineNumber, uint8_t byteToWrite) {
// char location expected in the fn args has been extrapolated from the pixel location
// format (see above), so chars go from 1 to LCDWIDTH/8, not from 0
  if((byteNumber <= LCDWIDTH/8) && (byteNumber != 0) && (lineNumber <=LCDHEIGHT) & (lineNumber != 0)) {
    byteNumber -= 1;
    lineNumber -= 1;
    frameBuffer[(lineNumber*LCDWIDTH/8)+byteNumber] = byteToWrite;
  } 
}


void setFrameBufferBlack() {
  for(uint16_t i=0; i < LCDWIDTH*LCDHEIGHT/8; i++) {
    frameBuffer[i] = 0x00;
  }
}


void setFrameBufferWhite() {
  for(uint16_t i=0; i<LCDWIDTH*LCDHEIGHT/8; i++) {
    frameBuffer[i] = 0xFF;
  }
}

void setFrameBufferWith(uint8_t data) {
	for(uint16_t i=0; i<LCDWIDTH*LCDHEIGHT/8; i++) {
		frameBuffer[i] = data;
	}
}


void drawImage(uint8_t *picData){
	for(uint16_t i = 0; i < LCDHEIGHT*LCDWIDTH/8; i++) {
		frameBuffer[i] = picData[i];
	}
}

void writeFrameBufferToDisplay() {
	writeMultipleLinesToDisplay(1, LCDHEIGHT, frameBuffer);
}


void clearLineBuffer() {
	setLineBufferWhite();
}


void clearFrameBuffer() {
	setFrameBufferWhite();
}


void clearDisplay() { 
	CS(GPIO_HI);
	delayMicroseconds(SCS_HIGH_DELAY);

	spi_writenb(&clearByte, 1);
	spi_writenb(&paddingByte, 1);
	delayMicroseconds(SCS_LOW_DELAY);

	CS(GPIO_LO);
	delayMicroseconds(INTERFRAME_DELAY);
}


// won't work if DISP pin is not used
void turnOff() {
	DISP(GPIO_HI);
}


// won't work if DISP pin is not used
void turnOn() {
	DISP(GPIO_LO);
}

unsigned int getDisplayWidth() {
  return LCDWIDTH;
}

unsigned int getDisplayHeight() {
  return LCDHEIGHT;
}

void hardToggleVcom(void) {
	static unsigned char vcom = 0;
	if (vcom == 0)
		ECOM(GPIO_HI);
	else
		ECOM(GPIO_HI);
	vcom = !vcom;
}

uint8_t reverseByte(uint8_t b) {
  b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
  b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
  b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
  return b;
}
